from abc import ABC, abstractmethod
from itertools import combinations
from typing import Annotated, Iterable

import pandas as pd


def _get_matching_subjects_by_slot(
    availabilities: pd.DataFrame,
    availabilities_column: str,
    availability_subject_column: str,
) -> pd.DataFrame:
    matching_subjects_by_slot = (
        availabilities.groupby(availabilities_column)
        .agg({availability_subject_column: lambda x: tuple(x)})
        .reset_index(col_level=1)
    )
    matching_subjects_by_slot = matching_subjects_by_slot[
        matching_subjects_by_slot[availability_subject_column].apply(len) > 1
    ]
    return matching_subjects_by_slot


def _get_matches_from_slots(
    matching_subjects_by_slot: pd.DataFrame,
    availabilities_column: str,
    availability_subject_column: str,
) -> pd.DataFrame:
    matches = (
        matching_subjects_by_slot.set_index(availabilities_column)[
            availability_subject_column
        ]
        .apply(lambda a: list(combinations(a, 2)))
        .reset_index()
        .explode(availability_subject_column)
    )
    matches[availability_subject_column] = matches[availability_subject_column].apply(
        lambda x: tuple(sorted(x))
    )
    return matches


def _apply_exclusions(
    matches: pd.DataFrame,
    exclusions: pd.DataFrame,
    availability_subject_column: str,
    exclusions_subject_columns: Annotated[Iterable[str], 2],
) -> pd.DataFrame:
    exclude_matchings = exclusions[exclusions_subject_columns]
    exclude_matchings = [tuple(sorted(x)) for x in exclude_matchings.values]
    exluded_mask = matches[availability_subject_column].isin(exclude_matchings)
    matches = matches.loc[~exluded_mask]
    return matches


class BackendBase(ABC):
    subject_features: list[str]
    availability_subject_column: str
    availabilities_column: str
    exclusions_subject_columns: Annotated[Iterable[str], 2]

    @abstractmethod
    def get_subjects(self, *args, **kwargs) -> pd.DataFrame:
        raise NotImplementedError()

    @abstractmethod
    def get_availabilities(self, *args, **kwargs) -> pd.DataFrame:
        raise NotImplementedError()

    @abstractmethod
    def get_exclusions(self, *args, **kwargs) -> pd.DataFrame:
        raise NotImplementedError()

    def get_all_matches(
        self,
        availabilities: pd.DataFrame | None = None,
        exclusions: pd.DataFrame | None | bool = None,
        return_matching_subjects_by_slot: bool = False,
    ) -> pd.DataFrame | tuple[pd.DataFrame, pd.DataFrame]:
        # initialize things
        if not availabilities:
            availabilities = self.get_availabilities()
        if isinstance(exclusions, bool) and exclusions:
            exclusions = self.get_exclusions()

        matching_subjects_by_slot = _get_matching_subjects_by_slot(
            availabilities,
            availabilities_column=self.availabilities_column,
            availability_subject_column=self.availability_subject_column,
        )
        matches = _get_matches_from_slots(
            matching_subjects_by_slot,
            availabilities_column=self.availabilities_column,
            availability_subject_column=self.availability_subject_column,
        )

        if isinstance(exclusions, pd.DataFrame):
            matches = _apply_exclusions(
                matches,
                exclusions=exclusions,
                availability_subject_column=self.availability_subject_column,
                exclusions_subject_columns=self.exclusions_subject_columns,
            )

        if return_matching_subjects_by_slot:
            return matches, matching_subjects_by_slot
        else:
            return matches



import pandas as pd
from swapanything.backend import _base


def test__get_matching_subjects_by_slot():
    AVAIL_COL = "avail"
    SUBJ_COL = "subj"
    df = pd.DataFrame(
        [
            ["sub1", "A"],
            ["sub2", "B"],
            ["sub5", "A"],
            ["sub3", "A"],
            ["sub4", "C"],
        ],
        columns=[SUBJ_COL, AVAIL_COL],
    )

    expected_result = pd.DataFrame(
        [
            ["A", ("sub1", "sub5", "sub3")],
        ],
        columns=[AVAIL_COL, SUBJ_COL],
    )

    result = _base._get_matching_subjects_by_slot(
        availabilities=df,
        availability_subject_column=SUBJ_COL,
        availabilities_column=AVAIL_COL,
    )
    assert isinstance(result, pd.DataFrame)
    assert result.equals(expected_result)


def test__get_matches_from_slots():
    AVAIL_COL = "avail"
    SUBJ_COL = "subj"
    slots = pd.DataFrame(
        [
            ["A", ("sub1", "sub5", "sub3")],
        ],
        columns=[AVAIL_COL, SUBJ_COL],
    )

    expected_result = pd.DataFrame(
        [
            ["A", ("sub1", "sub5")],
            ["A", ("sub1", "sub3")],
            ["A", ("sub3", "sub5")],
        ],
        columns=[AVAIL_COL, SUBJ_COL],
        index=[0, 0, 0],
    )

    result = _base._get_matches_from_slots(
        slots, availabilities_column=AVAIL_COL, availability_subject_column=SUBJ_COL
    )
    assert isinstance(result, pd.DataFrame)
    assert result.equals(expected_result)


def test__apply_exclusions():
    SUBJ_COL = "subj"
    EXCL_SUBJ_COLS = ["s1", "s2"]

    matches = pd.DataFrame(
        [
            ["A", ("sub1", "sub5")],
            ["A", ("sub1", "sub3")],
            ["A", ("sub3", "sub5")],
        ],
        columns=["avail", SUBJ_COL],
        index=[0, 0, 0],
    )
    exclusions = pd.DataFrame(
        [
            ["e1", "sub1", "sub3"],
        ],
        columns=["excl", EXCL_SUBJ_COLS[0], EXCL_SUBJ_COLS[1]],
    )
    expected_result = matches.loc[[True, False, True]]

    result = _base._apply_exclusions(
        matches,
        exclusions,
        availability_subject_column=SUBJ_COL,
        exclusions_subject_columns=EXCL_SUBJ_COLS,
    )
    assert isinstance(result, pd.DataFrame)
    assert result.equals(expected_result)


def test_base_backend():
    AVAIL_COL = "avail"
    SUBJ_COL = "subj"
    EXCL_SUBJ_COLS = ["s1", "s2"]

    class TestBackend(_base.BackendBase):
        def __init__(self) -> None:
            self.subject_features = ["a", "b", "c"]
            self.availability_subject_column = SUBJ_COL
            self.availabilities_column = AVAIL_COL
            self.exclusions_subject_columns = EXCL_SUBJ_COLS

        def get_subjects(self) -> None:
            return

        def get_availabilities(self) -> pd.DataFrame:
            return pd.DataFrame(
                [
                    ["sub1", "A"],
                    ["sub2", "B"],
                    ["sub5", "A"],
                    ["sub3", "A"],
                    ["sub4", "C"],
                ],
                columns=[SUBJ_COL, AVAIL_COL],
            )

        def get_exclusions(self) -> pd.DataFrame:
            return pd.DataFrame(
                [
                    ["e1", "sub1", "sub3"],
                ],
                columns=["excl", EXCL_SUBJ_COLS[0], EXCL_SUBJ_COLS[1]],
            )

    expected_result = pd.DataFrame(
        [
            ["A", ("sub1", "sub5")],
            ["A", ("sub3", "sub5")],
        ],
        columns=["avail", SUBJ_COL],
        index=[0, 0],
    )

    be = TestBackend()
    result = be.get_all_matches(exclusions=True)
    assert isinstance(result, pd.DataFrame)
    assert result.equals(expected_result)
